
/*

Copyright (C) 2015 Dirk Schuetz <dirk.schuetz@durham.ac.uk> and Lewin Strauss

    This file is part of KnotJob.

    KnotJob is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    KnotJob is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/


import java.util.ArrayList;
import java.util.Arrays;

/**
 * A Library of operations on integer matrices.
 */
public class IntegerMatrices {
    
    /** 
     * Diagonalises a mod4 matrix.
     * Tracks either colOp (!trackInv) or colOpInv (trackInv).
     */
    public static ArrayList<int[][]> trackedDiag_4(int[][] matrix, boolean trackInv) { 
        int l = matrix.length;
        int m = matrix[0].length;
        int[][] diag = new int[l][m];
        int[][] trackedMat = new int[m][m];
        for (int i = 0; i < m; i++) trackedMat[i][i] = 1;
        
        for (int i = 0; i < BasicMaths.min(l, m); i++) { 
            int[][] zeroMat = new int[l - i][m - i];
            if (Arrays.deepEquals(matrix, zeroMat)) break;
            
            int[][] redTrMat = trackedDiagStep_4(matrix, trackInv);
            diag[i][i] = matrix[0][0]; // read off diag rather than using matMult
            matrix = getSubMat(matrix, 1);
            if (i > 0) {                                                  
                if (trackInv) {
                    int[][] colOpInv = getSuperMat(redTrMat, i);  
                    trackedMat = mult_n(colOpInv, trackedMat, 4);
                } else {
                    int[][] colOp = getSuperMat(redTrMat, i);  
                    trackedMat = mult_n(trackedMat, colOp, 4);
                } 
            } else trackedMat = redTrMat;
        }
        
        ArrayList<int[][]> resultList = new ArrayList<int[][]>(2);
        resultList.add(diag);
        resultList.add(trackedMat);
        return resultList;
    }
    
    /**
     * Diagonalises a mod4 matrix of the form X|2Y 
     * w/o tampering w/ the boundary representation.
     */
    public static int[][] cutDiag(int[][] matrix, int cut) {
        int l = matrix.length;
        int m = matrix[0].length;
        
        if (cut > 0) {
            int[][] matrixTop = new int[cut][m];
            for (int i = 0; i < cut; i++) {
                for (int j = 0; j < m; j++) {
                    matrixTop[i][j] = matrix[i][j];
                }
            }
            int[][] matrixBot = new int[l - cut][m];
            for (int i = 0; i < (l - cut); i++) {
                for (int j = 0; j < m; j++) {
                    matrixBot[i][j] = matrix[i + cut][j];
                }
            }
            
            ArrayList<int[][]> topTrDiagResult = trackedDiag_4(matrixTop, false);
            matrixTop = topTrDiagResult.get(0);
            int[][] colOp = topTrDiagResult.get(1);
            matrixBot = mult_n(matrixBot, colOp, 4);
            for (int i = 0; i < BasicMaths.min(cut, m); i++) {
                matrix[i][i] = matrixTop[i][i];
            }
            for (int i = cut; i < l; i++) {
                for (int j = 0; j < m; j++) {
                    matrix[i][j] = matrixBot[i - cut][j];
                }
            }
        }
        return trackedDiag_4(matrix, false).get(0);
    }
    
    /** 
     * Performs one step in the diagonalisation process of mod4 matrices.
     * Tracks either colOp matrix (!trackInv) or colOpInv matrix (trackInv).
     */
    public static int[][] trackedDiagStep_4(int[][] matrix, boolean trackInv) {
        boolean two = false;
        int i_pivot = -1;
        int j_pivot = -1;
        int l = matrix.length;
        int m = matrix[0].length;
        
        int[][] trackedOp = new int[m][m];
        for (int i = 0; i < m; i++) trackedOp[i][i] = 1;
        
        search: {
            for (int j = 0; j < m; j++) {
                for (int i = 0; i < l; i++) {
                    if (matrix[i][j] == 1 || matrix[i][j] == 3) {
                        i_pivot = i;
                        j_pivot = j;
                        break search;
                    }
                }
            }
            for (int j = 0; j < m; j++) {
                for (int i = 0; i < l; i++) {
                    if (matrix[i][j] == 2) {
                        i_pivot = i;
                        j_pivot = j;
                        two = true;
                        break search;
                    }
                }
            }
        }
               
        if (matrix[i_pivot][j_pivot] == 3) {
            for (int j = 0; j < matrix[0].length; j++) {
                matrix[i_pivot][j] = BasicMaths.mod_n(3 * matrix[i_pivot][j], 4); 
            }
        }
        if (j_pivot > 0) {                                         
            swapCol(matrix, 0, j_pivot);
            swapCol(trackedOp, 0, j_pivot);
        }
        swapRow(matrix, 0, i_pivot);
        for (int i = 1; i < l; i++) { 
            if (matrix[i][0] != 0) {
                if (!two) addRow_n(matrix, (-1 * matrix[i][0]), 0, i, 4);
                else addRow_n(matrix, -1, 0, i, 4);
            }
        }
        for (int j = 1; j < m; j++) { 
            if (matrix[0][j] != 0) { 
                int holder = matrix[0][j];
                if (!two) addCol_n(matrix, -1 * matrix[0][j], 0, j, 4);
                else addCol_n(matrix, -1, 0, j, 4);
                int[][] trMat = new int[m][m]; 
                for (int i = 0; i < m; i++) trMat[i][i] = 1;
                if (!trackInv) {
                    if (!two) trMat[0][j] = -1*holder;
                    else trMat[0][j] = -1;
                    trackedOp = mult_n(trackedOp, trMat, 4);
                } else {
                    if (!two) trMat[0][j] = holder;    
                    else trMat[0][j] = 1; 
                    trackedOp = mult_n(trMat, trackedOp, 4);
                }
            }
        }
        
        return trackedOp;
    } 
    
    /**
     * Adds (factorA * matrix a) to (factorB * matrix b) considering modulo (=mod) arithmetic.
     */
    public static int[][] add_n(int factorA, int[][] a, int factorB, int[][] b, int mod) {
        int[][] result = new int[a.length][b[0].length];
        for (int i = 0; i <  a.length; i++) {
            for (int j = 0; j< b[0].length; j++) {
                result[i][j] = BasicMaths.mod_n((factorA * a[i][j]) + (factorB * b[i][j]), mod);
            }
        }
        return result;
    }
    
    /**
     * Multiplies two matrices a and b considering modulo (=mod) arithmetic. 
     */
    public static int[][] mult_n(int[][] a, int[][] b, int mod) {   
        int[][] result = new int[a.length][b[0].length];
        for (int i = 0; i < a.length; i++) {
            for (int j = 0; j < b[0].length; j++) {
                for (int k = 0; k < a[0].length; k++) {
                    result[i][j] += a[i][k] * b[k][j];
                }
                result[i][j] = BasicMaths.mod_n(result[i][j], mod);
            }
        }
        return result;
    }
    
    /**
     * Adds add-th row to addTo-th row considering modulo (=mod) arithmetic. 
     */
    public static void addRow_n(int[][] matrix, int factor , int add, int addTo, int mod) {
        for (int j = 0; j < matrix[0].length; j++) {
            matrix[addTo][j] = BasicMaths.mod_n(factor * matrix[add][j] + matrix[addTo][j], mod);
        }
    }
    
    /**
     * Adds add-th column to addTo-th column, considering modulo (=mod) arithmetic.
     */
    public static void addCol_n(int[][] matrix, int factor , int add, int addTo, int mod) {
        for (int i = 0; i < matrix.length; i++) {
            matrix[i][addTo] = BasicMaths.mod_n(factor * matrix[i][add] + matrix[i][addTo], mod);
        }
    }
    
    /**
     * Swaps a-th and b-th row in a matrix.
     */
    public static void swapRow(int[][] matrix, int a, int b) { 
        for (int j = 0; j < matrix[0].length; j++) {
            int holder = matrix[b][j];
            matrix[b][j] = matrix[a][j];
            matrix[a][j] = holder;
        }
    }
    
    /**
     * Swaps a-th and b-th column in a matrix.
     */
    public static void swapCol(int[][] matrix, int a, int b) {  
        for (int i = 0; i < matrix.length; i++) {
            int holder = matrix[i][b];
            matrix[i][b] = matrix[i][a];
            matrix[i][a] = holder;
        }
     }
    
    /**
     * Reduces matrix from the left/top by "= reduceBy" rows + columns.
     */
    public static int[][] getSubMat(int[][] matrix, int reduceBy) {     
        int numRows = matrix.length;
        int numCols = matrix[0].length;
        int[][] sub = new int[numRows - reduceBy][numCols - reduceBy];
        for (int i = 0; i < (numRows - reduceBy); i++) {
            for (int j = 0; j < (numCols - reduceBy); j++) {
                sub[i][j] = matrix[reduceBy + i][reduceBy + j];
            }
        }
        return sub;
    }
    
    /**
     * Extends matrix to the left/top with "= extendBy" identity rows + columns.
     */
    public static int[][] getSuperMat(int[][] matrix, int extendBy) {
        int numRows = matrix.length;
        int numCols = matrix[0].length;
        int[][] sup = new int[numRows + extendBy][numCols + extendBy];
        for (int i = 0; i < extendBy; i++) sup[i][i] = 1;
        for (int i = extendBy; i < (numRows + extendBy); i++) {
            for (int j = extendBy; j < (numCols + extendBy); j++) {
                sup[i][j] = matrix[i - extendBy][j - extendBy];
            }
        }
        return sup;
    }
    
    
    public static void main(String[] args) {
        
        int[][] myArray = { {1,2,3,1,2,0},{1,2,1,2,1,2},{2,2,2,2,0,0},{2,2,2,2,2,2},{2,2,2,2,2,2},{2,0,0,0,2,2} };
        
        
        
        int[][] result = cutDiag(myArray, 3);

        
        
    }
}

